﻿using System.Threading;
using System.Windows.Media.Media3D;
using ProjectorToolkit;
using ProjectorToolkit.Elements;
using ProjectorToolkit.Relations;

namespace Pointing_and_CrossingRelation
{
    /// <summary>
    /// Changes some image if it is hovered, i.e. if the projector points at it.
    /// Entering and leaving the image in under 1 second toggles another picture element.
    /// </summary>
    class Program
    {
        private IEnvironment env;

        // Timer to remove crossing notification
        private Timer timer;
        // Notification image that is shown after crossing the image with the projector
        ImageElement crossed = new ImageElement(new Point3D(0, 1100, -198), 200, 80, @"images\crossed.png");

        static void Main(string[] args)
        {
            new Program();
        }

        public Program()
        {
            // Create a new environment
            env = new XmlEnvironment();

            // Add an image to the world. Later, this image changes its picture while pointing.
            OrthographicDisplay smartboard = (OrthographicDisplay)env.World.Get("Smartboard");
            ImageElement image = new ImageElement(640, 480, @"images\1.jpeg");
            image.PositionOn(smartboard, smartboard.ResolutionX / 2, smartboard.ResolutionY / 2);
            env.World.Add(image);

            // Create a PointingRelation between the projector and the image
            // Started changes the image source and ended reverts it back
            AbstractElement projector = env.World.Get("Projector");
            PointingRelation pr = new PointingRelation(projector, image);
            pr.PointingStarted += PointingStart;
            pr.PointingEnded += PointingEnd;

            // Create a CrossingRelation between the projector and the image
            // When the image is crossed in less than a second, the crossed picture appears as a notification.
            CrossingRelation cr = new CrossingRelation(projector, image, 1.0);
            cr.Cutted += Cutted;

            // Register a cursor
            env.CursorDB.Register(projector);
        }

        private void Cutted(object sender, CrossingRelationEventArgs e)
        {
            // If timer does not exist, ie notification is invisible
            if (timer == null)
            {
                // Add image to the world and remove it after 3 seconds
                env.World.Add(crossed);
                timer = new Timer(RemoveCutting, crossed, 3000, 0);
            }
            else
            {
                // Show the notification for another 3 seconds
                timer.Change(3000, 0);
            }
        }

        private void RemoveCutting(object state)
        {
            // Destroy timer and remove notification
            timer.Dispose();
            env.World.Remove((AbstractElement)state);
            timer = null;
        }

        private void PointingStart(object sender, PointingRelationEventArgs e)
        {
            // Change image picture while pointing at it
            ((ImageElement)e.Target).Picture = @"images\2.jpeg";
        }

        private void PointingEnd(object sender, PointingRelationEventArgs e)
        {
            // Change image picture back to the original picture
            ((ImageElement)e.Target).Picture = @"images\1.jpeg";
        }
    }
}