﻿using System;
using System.Windows;
using System.Windows.Media.Media3D;
using GroupLab.iNetwork;
using ProjectorToolkit.Utilities;

namespace ProjectorToolkit.Elements
{
    public class PerspectiveDisplay : AbstractDisplay
    {
        private Angle horizontalAngle;
        protected int centerPixelY;
        protected double pixelStep;
        protected int halfResolutionX, halfResolutionY;
        private Vector3D planeUp, planeRight;
        #region Properties
        public Angle HorizontalAngle
        {
            get { return horizontalAngle; }
        }
        #endregion

        public PerspectiveDisplay() : base() { }
        public PerspectiveDisplay(int resolutionX, int resolutionY, Angle horizontalAngle)
            : base(resolutionX, resolutionY)
        {
            Initialize(horizontalAngle, ResolutionY / 2);
        }
        public PerspectiveDisplay(int resolutionX, int resolutionY, Angle horizontalAngle, int centerPixelY)
            : base(resolutionX, resolutionY)
        {
            Initialize(horizontalAngle, centerPixelY);
        }
        public void Initialize(Angle horizontalAngle, int centerPixelY)
        {
            this.horizontalAngle = horizontalAngle;
            this.centerPixelY = centerPixelY;

            halfResolutionX = (int)Math.Ceiling(ResolutionX / 2.0);
            halfResolutionY = (int)Math.Ceiling(ResolutionY / 2.0);

            pixelStep = Math.Tan(horizontalAngle.InRadians / 2) / Math.Ceiling(ResolutionX / 2.0);
        }

        public override Camera Camera
        {
            get
            {
                return new PerspectiveCamera(Position, Direction, Up, HorizontalAngle.InDegrees);
            }
        }

        public override void SetPose(Point3D position, Vector3D direction, Vector3D up)
        {
            // Save plane direction to calculate pixel directions
            planeUp = up;
            planeRight = Vector3D.CrossProduct(direction, up);

            // Move camera to represent center ray
            Vector3D colDiff = pixelStep * (halfResolutionY - centerPixelY) * up;
            Vector3D cDir = direction - colDiff;
            cDir.Normalize();
            direction.Normalize();
            Vector3D cUp = Vector3DExtensions.SameRotation(up, direction, cDir);
            cUp.Normalize();

            base.SetPose(position, cDir, cUp);
        }

        public override Vector3D PixelDirection(int x, int y)
        {
            Vector3D rowDiff = pixelStep * (x - halfResolutionX) * planeRight;
            Vector3D colDiff = pixelStep * (y - halfResolutionY) * planeUp;
            Vector3D result = Direction + rowDiff - colDiff;
            result.Normalize();

            return result;
        }

        public override Point3D PixelOrigin(int x, int y)
        {
            return Position;
        }

        public override Point ConvertToPixel(Point3D point)
        {
            Vector3D direction = point - Position;
            direction.Normalize();
            Vector3D deviance = direction - Direction;

            int x = (int)Math.Round(Vector3D.DotProduct(deviance, Right) / pixelStep + halfResolutionX);
            int y = (int)Math.Round(Vector3D.DotProduct(deviance, Up) / pixelStep + halfResolutionY);
            return new Point(x, ResolutionY - 1 - y);
        }

        #region Synchronize
        public override void UpdateElement(Message msg)
        {
            horizontalAngle = Angle.FromDegrees(msg.GetDoubleField("horizontal_angle"));
            planeRight = msg.GetVector3D("plane_right");
            planeUp = msg.GetVector3D("plane_up");
            base.UpdateElement(msg);
            Initialize(horizontalAngle, ResolutionY / 2);
        }

        public override Message ToMessage(string messageName)
        {
            Message msg = base.ToMessage(messageName);
            msg.AddField("horizontal_angle", horizontalAngle.InDegrees);
            msg.Add("plane_right", planeRight);
            msg.Add("plane_up", planeUp);
            return msg;
        }
        #endregion
    }
}
