﻿using System.Windows.Media.Media3D;
using ProjectorToolkit.Utilities;

namespace ProjectorToolkit.Elements
{
    /// <summary>
    /// The origin of a 3D space as a singleton implementation. The direction
    /// equals the z-axis, the up-vector the y-axis. The usage of this element
    /// is to simulate absolute positioning by using it as relative anchor.
    /// </summary>
    sealed class Origin : AbstractElement
    {
        private static Origin instance;

        public static Origin Instance
        {
            get
            {
                if (instance == null)
                    instance = new Origin();
                return instance;
            }
        }

        private Origin()
        {
            SetPose(new Point3D(0, 0, 0), new Vector3D(0, 0, 1), new Vector3D(0, 1, 0));
        }
    }

    /// <summary>
    /// This class extends the base Element with relative or absolute 
    /// positioning of the element.
    /// </summary>
    public abstract class RelativeElement : AbstractElement
    {
        private AbstractElement relativeElement = Origin.Instance;

        #region Relative to Absolute
        public static Point3D RelativeToAbsolute(Point3D point, IPositionable origin)
        {
            return ((point.X * -origin.Right) +
                    (point.Y * origin.Up) +
                    (point.Z * origin.Direction)).ToPoint3D();
        }

        public static Vector3D RelativeToAbsolute(Vector3D vector, IPositionable origin)
        {
            return (vector.X * -origin.Right) +
                   (vector.Y * origin.Up) +
                   (vector.Z * origin.Direction);
        }
        #endregion
        #region Relative to absolute properties
        public override Point3D Position
        {
            get
            {
                return relativeElement.Position +
                    RelativeToAbsolute(base.Position, relativeElement).ToVector3D();
            }
        }

        public override Vector3D Direction
        {
            get
            {
                return RelativeToAbsolute(base.Direction, relativeElement);
            }
        }

        public override Vector3D Up
        {
            get
            {
                return RelativeToAbsolute(base.Up, relativeElement);
            }
        }
        #endregion

        public void SetRelativeTo(AbstractElement element)
        {
            ResetRelativeElement(element);
        }

        public void SetAbsolute()
        {
            ResetRelativeElement(Origin.Instance);
        }

        private void ResetRelativeElement(AbstractElement element)
        {
            relativeElement.Updated -= RelativeElementUpdated;
            relativeElement = element;
            relativeElement.Updated += RelativeElementUpdated;
            UpdateNotification();
        }

        private void RelativeElementUpdated(AbstractElement element)
        {
            UpdateNotification();
        }
    }
}
