﻿using System;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Media3D;
using GroupLab.iNetwork;
using ProjectorToolkit.Utilities;

namespace ProjectorToolkit.Elements
{
    public abstract class AbstractVirtualElement2D : RelativeElement, IPositionable2D
    {
        private GeometryModel3D mesh;
        private MeshGeometry3D geometry;
        private Polygon corners = new Polygon(new Point[0]);
        private bool blackout = false;
        #region Properties
        /// <summary>
        /// Edges of the 2D element. These are the edges around the position of the element.
        /// Position, up and direction will change those coordinates.
        /// </summary>
        public virtual Polygon Corners
        {
            get { return corners; }
            set
            {
                if (value == null)
                    corners = new Polygon(new Point[0]);
                else if (value.Length < 3)
                    throw new ArgumentException("A 2D element needs at least 3 edges!");
                else
                    corners = value;
                UpdateNotification();
            }
        }

        public override Model3D Mesh
        {
            get
            {
                if (mesh == null)
                    CreateMesh();
                return mesh;
            }
        }

        /// <summary>
        /// Blacks out the element. The element is not getting transparent,
        /// all underlying elements are invisible.
        /// </summary>
        public bool Blackout
        {
            get { return blackout; }
            set
            {
                if (value != blackout)
                {
                    blackout = value;
                    UpdateNotification();
                }
            }
        }
        #endregion

        #region Constructor
        public AbstractVirtualElement2D() : base() { }
        public AbstractVirtualElement2D(double width, double height)
            : base()
        {
            SetRectangle(width, height);
        }
        public AbstractVirtualElement2D(Point3D position, double width, double height)
            : this(width, height)
        {
            SetPosition(position);
        }

        private void CreateMesh()
        {
            mesh = new GeometryModel3D();
            geometry = new MeshGeometry3D();

            PointCollection texture = new PointCollection();
            texture.Add(new Point(1, 0));
            texture.Add(new Point(1, 1));
            texture.Add(new Point(0, 1));
            texture.Add(new Point(0, 0));
            geometry.TextureCoordinates = texture;

            mesh.Material = new DiffuseMaterial(new SolidColorBrush(Colors.Red));
            mesh.Geometry = geometry;
        }
        #endregion

        protected abstract Material Texture();

        public override void UpdateMesh()
        {
            base.UpdateMesh();
            Point3DCollection positions = new Point3DCollection(4);
            for (int i = 0; i < corners.Length; i++)
                positions.Add(Position + corners[i].X * Right + Corners[i].Y * Up);
            geometry.Positions = positions;

            Vector3DCollection normals = new Vector3DCollection(corners.Length);
            foreach (Point3D p in geometry.Positions)
                normals.Add(Direction);
            geometry.Normals = normals;

            Int32Collection indices = new Int32Collection();
            for (int i = 1; i + 1 < geometry.Positions.Count; i++)
            {
                indices.Add(0);
                indices.Add(i + 1);
                indices.Add(i);
            }
            indices.Add(geometry.Positions.Count - 2);
            indices.Add(geometry.Positions.Count - 1);
            indices.Add(0);
            geometry.TriangleIndices = indices;

            if (blackout)
            {
                DiffuseMaterial material = new DiffuseMaterial();
                material.Brush = new SolidColorBrush(Colors.Black);
                mesh.Material = material;
            }
            else
                mesh.Material = Texture();
        }

        public override Point3D? Intersection(Point3D origin, Vector3D direction)
        {
            return (new Polygon3D(Position, Direction, Corners)).Intersection(origin, direction);
        }

        public void SetRectangle(double width, double height)
        {
            double hwidth = width / 2;
            double hheight = height / 2;
            Corners = new Polygon(new Point[] {
                    new Point(-hwidth, +hheight),
                    new Point(-hwidth, -hheight),
                    new Point(+hwidth, -hheight),
                    new Point(+hwidth, +hheight)
                });
        }

        #region Synchronization
        public override Message ToMessage(string messageName)
        {
            Message msg = base.ToMessage(messageName);
            msg.AddField("edge_count", corners.Length);

            for (int i = 0; i < corners.Length; i++)
            {
                msg.AddField("edge_x" + i, corners.Points[i].X);
                msg.AddField("edge_y" + i, corners.Points[i].Y);
            }

            return msg;
        }

        public override void UpdateElement(Message msg)
        {
            int edgeCount = msg.GetIntField("edge_count");

            corners = new Polygon(new Point[edgeCount]);
            for (int i = 0; i < 4; i++)
            {
                corners.Points[i] = new Point(
                    msg.GetDoubleField("edge_x" + i),
                    msg.GetDoubleField("edge_y" + i)
                );
            }
            base.UpdateElement(msg);
        }
        #endregion
    }
}
