﻿using System.Collections.Generic;
using System.Diagnostics;
using System.Net;
using GroupLab.iNetwork;
using GroupLab.iNetwork.Tcp;

namespace ProjectorToolkit.Network
{
    /// <summary>
    /// A client with buffer functionality. It buffers all send messages that 
    /// are sent before a connection is established. It also buffers incoming 
    /// and outgoing messages if Buffering while set to true. To start sending
    /// and receiving messages, set Buffering to false.
    /// </summary>
    public class BufferClient
    {
        private bool buffer = true;
        private Connection connection;
        private List<Message> bufferOutMessages = new List<Message>();
        private List<Message> bufferInMessages = new List<Message>();

        public delegate void MessageEventHandler(object sender, Message msg);
        public event MessageEventHandler MessageReceived = delegate { };

        #region Properties
        public bool Buffering
        {
            get { return buffer; }
            set
            {
                buffer = value;
                if (buffer)
                {
                    ReceiveAllBufferedMessages();
                    SendAllBufferedMessages();
                }
            }
        }
        #endregion

        protected virtual void OnMessageReceived(object sender, Message msg)
        {
            Debug.Assert(msg != null);

            if (buffer)
                bufferInMessages.Add(msg);
            else
                MessageReceived(sender, msg);
        }

        public void Send(Message msg)
        {
            Debug.Assert(msg != null);

            if (connection == null || buffer)
                bufferInMessages.Add(msg);
            else
                connection.SendMessage(msg);
        }

        #region BufferClient
        public BufferClient()
        {
            Connector connector = new Connector();
            connector.SingleConnectionDiscovered += new SingleConnectionDiscoveryEventHandler(OnConnectorDiscovered);
            connector.Discover();
        }

        public BufferClient(IPAddress ip, int port)
        {
            OnConnectorDiscovered(new Connection(ip, port));
        }

        ~BufferClient()
        {
            ShutDown();
        }

        public void ShutDown()
        {
            if (connection != null)
                connection.Stop();
        }

        private void OnConnectorDiscovered(Connection connection)
        {
            this.connection = connection;
            if (connection != null)
            {
                connection.MessageReceived += OnMessageReceived;
                connection.Start();

                if (!buffer)
                    SendAllBufferedMessages();
            }
        }

        private void ReceiveAllBufferedMessages()
        {
            // Send all buffered messages
            while (bufferOutMessages.Count > 0)
            {
                Send(bufferOutMessages[0]);
                bufferOutMessages.RemoveAt(0);
            }
        }

        private void SendAllBufferedMessages()
        {
            // Send all buffered messages
            while (bufferInMessages.Count > 0)
            {
                MessageReceived(this, bufferInMessages[0]);
                bufferInMessages.RemoveAt(0);
            }
        }
        #endregion
    }
}
