﻿using System.Collections.Generic;
using System.Windows.Media.Media3D;
using ProjectorToolkit.Elements;
using ProjectorToolkit.Utilities;

namespace ProjectorToolkit.Relations
{
    public class OverlappingElementRelationEventArgs : OverlappingRelationEventArgs
    {
        public AbstractDisplay Display;
        public DisplayArgs DisplayArgs;
        public AbstractVirtualElement2D Element;
        public ElementArgs ElementArgs;

        public OverlappingElementRelationEventArgs(Overlap overlapping) : base(overlapping) { }
    }

    /// <summary>
    /// This relation triggers callbacks when a display overlaps an element.
    /// </summary>
    public sealed class OverlappingDisplayElementRelation : AbstractOverlappingRelation<OverlappingElementRelationEventArgs>
    {
        public delegate void OverlappingElementEventHandler(object sender, OverlappingElementRelationEventArgs e);
        public event OverlappingElementEventHandler Overlapped = delegate { };
        public event OverlappingElementEventHandler OverlappingStarted = delegate { };
        public event OverlappingElementEventHandler OverlappingMoved = delegate { };
        public event OverlappingElementEventHandler OverlappingEnded = delegate { };
        private AbstractDisplay display;
        private AbstractVirtualElement2D element;
        private Polygon area;

        #region Constructors
        public OverlappingDisplayElementRelation(AbstractDisplay display, AbstractVirtualElement2D element)
            : this(display, display.PixelArea, element) { }

        /// <summary>
        /// Create an OverlappingRelation between a display and an element.
        /// If no area is specified, the whole viewport is used.
        /// </summary>
        /// <param name="display">The display of the overlapping</param>
        /// <param name="area">[Optional] Polygon of pixels specifying the area in the viewport 1</param>
        /// <param name="element">The element of the overlapping</param>
        public OverlappingDisplayElementRelation(AbstractDisplay display, Polygon area, AbstractVirtualElement2D element)
        {
            this.display = display;
            this.area = area;
            this.element = element;

            display.Updated += NoRecursiveUpdate;
            element.Updated += NoRecursiveUpdate;
        }
        #endregion

        protected override void Update(AbstractElement element)
        {
            Overlap overlapping = Overlap.Of(display, area, this.element);

            RelationEventType? type = SetRelationType(overlapping != null);
            if (type != null)
            {
                OverlappingElementRelationEventArgs args = CreateEventArgs((RelationEventType)type, overlapping);

                if (type == RelationEventType.Start)
                    OverlappingStarted(this, args);
                if (type == RelationEventType.Move)
                    OverlappingMoved(this, args);
                if (type == RelationEventType.End)
                    OverlappingEnded(this, args);
                Overlapped(this, args);
            }
        }

        private OverlappingElementRelationEventArgs CreateEventArgs(RelationEventType eventType, Overlap overlapping)
        {
            OverlappingElementRelationEventArgs args = new OverlappingElementRelationEventArgs(overlapping);
            SetDefaultEventArgs(args, eventType);
            args.Display = display;
            args.Element = element;
            if (overlapping != null)
            {
                args.DisplayArgs = new DisplayArgs(display, area, overlapping);
                args.ElementArgs = new ElementArgs(element, overlapping);
            }
            return args;
        }
    }
}
