﻿using System.Collections.Generic;
using System.Windows.Media.Media3D;
using ProjectorToolkit.Elements;
using ProjectorToolkit.Utilities;

namespace ProjectorToolkit.Relations
{
    public class OverlappingDisplayRelationEventArgs : OverlappingRelationEventArgs
    {
        public AbstractDisplay Display1;
        public AbstractDisplay Display2;
        public DisplayArgs DisplayArgs1;
        public DisplayArgs DisplayArgs2;

        public OverlappingDisplayRelationEventArgs(Overlap overlapping) : base(overlapping) { }
    }

    /// <summary>
    /// This relation triggers callbacks when two displays overlap each other.
    /// </summary>
    public sealed class OverlappingDisplaysRelation : AbstractOverlappingRelation<OverlappingDisplayRelationEventArgs>
    {
        public delegate void OverlappingDisplayEventHandler(object sender, OverlappingDisplayRelationEventArgs e);
        public event OverlappingDisplayEventHandler Overlapped = delegate { };
        public event OverlappingDisplayEventHandler OverlappingStarted = delegate { };
        public event OverlappingDisplayEventHandler OverlappingMoved = delegate { };
        public event OverlappingDisplayEventHandler OverlappingEnded = delegate { };
        private AbstractDisplay display1;
        private AbstractDisplay display2;
        private Polygon area1;
        private Polygon area2;

        #region Constructors
        public OverlappingDisplaysRelation(AbstractDisplay display1, AbstractDisplay display2)
            : this(display1, display1.PixelArea, display2, display2.PixelArea) { }

        public OverlappingDisplaysRelation(AbstractDisplay display1, Polygon area1, AbstractDisplay display2)
            : this(display1, area1, display2, display2.PixelArea) { }

        public OverlappingDisplaysRelation(AbstractDisplay display1, AbstractDisplay display2, Polygon area2)
            : this(display1, display1.PixelArea, display2, area2) { }

        /// <summary>
        /// Create an OverlappingRelation between areas of viewports.
        /// If no area is specified, the whole viewport is used.
        /// </summary>
        /// <param name="display1">The first viewport of the overlapping</param>
        /// <param name="area1">[Optional] Polygon of pixels specifying the area in the viewport 1</param>
        /// <param name="display2">The second viewport of the overlapping</param>
        /// <param name="area2">[Optional] Polygon of pixels specifying the area in the viewport 2</param>
        public OverlappingDisplaysRelation(AbstractDisplay display1, Polygon area1, AbstractDisplay display2, Polygon area2)
        {
            this.display1 = display1;
            this.area1 = area1;
            this.display2 = display2;
            this.area2 = area2;

            display1.Updated += NoRecursiveUpdate;
            display2.Updated += NoRecursiveUpdate;
        }
        #endregion

        protected override void Update(AbstractElement element)
        {
            Overlap overlapping = Overlap.Of(display1, area1, display2, area2);

            RelationEventType? type = SetRelationType(overlapping != null);
            if (type != null)
            {
                OverlappingDisplayRelationEventArgs args = CreateEventArgs((RelationEventType)type, overlapping);

                if (type == RelationEventType.Start)
                    OverlappingStarted(this, args);
                if (type == RelationEventType.Move)
                    OverlappingMoved(this, args);
                if (type == RelationEventType.End)
                    OverlappingEnded(this, args);
                Overlapped(this, args);
            }
        }

        private OverlappingDisplayRelationEventArgs CreateEventArgs(RelationEventType eventType, Overlap overlapping)
        {
            OverlappingDisplayRelationEventArgs args = new OverlappingDisplayRelationEventArgs(overlapping);
            SetDefaultEventArgs(args, eventType);
            args.Display1 = display1;
            args.Display2 = display2;
            if (overlapping != null)
            {
                args.DisplayArgs1 = new DisplayArgs(display1, area1, overlapping);
                args.DisplayArgs2 = new DisplayArgs(display2, area2, overlapping);
            }
            return args;
        }
    }
}
