﻿using System.Collections.Generic;
using System.Windows.Media.Media3D;
using ProjectorToolkit.Elements;
using ProjectorToolkit.Utilities;

namespace ProjectorToolkit.Relations
{
    public class OverlappingElementsRelationEventArgs : OverlappingRelationEventArgs
    {
        public AbstractElement Element1;
        public AbstractElement Element2;
        public ElementArgs ElementArgs1;
        public ElementArgs ElementArgs2;

        public OverlappingElementsRelationEventArgs(Overlap overlapping) : base(overlapping) { }
    }

    /// <summary>
    /// This relation triggers callbacks when two elements overlap each other.
    /// </summary>
    public sealed class OverlappingElementsRelation : AbstractOverlappingRelation<OverlappingDisplayRelationEventArgs>
    {
        public delegate void OverlappingElementsEventHandler(object sender, OverlappingElementsRelationEventArgs e);
        public event OverlappingElementsEventHandler Overlapped = delegate { };
        public event OverlappingElementsEventHandler OverlappingStarted = delegate { };
        public event OverlappingElementsEventHandler OverlappingMoved = delegate { };
        public event OverlappingElementsEventHandler OverlappingEnded = delegate { };
        private AbstractVirtualElement2D element1;
        private AbstractVirtualElement2D element2;

        #region Constructors
        /// <summary>
        /// Create an OverlappingRelation between two elements.
        /// </summary>
        /// <param name="element1">The first element of the overlapping</param>
        /// <param name="element2">The second element of the overlapping</param>
        public OverlappingElementsRelation(AbstractVirtualElement2D element1, AbstractVirtualElement2D element2)
        {
            this.element1 = element1;
            this.element2 = element2;

            element1.Updated += NoRecursiveUpdate;
            element2.Updated += NoRecursiveUpdate;
        }
        #endregion

        protected override void Update(AbstractElement element)
        {
            Overlap overlapping = Overlap.Of(element1, element2);

            RelationEventType? type = SetRelationType(overlapping != null);
            if (type != null)
            {
                OverlappingElementsRelationEventArgs args = CreateEventArgs((RelationEventType)type, overlapping);

                if (type == RelationEventType.Start)
                    OverlappingStarted(this, args);
                if (type == RelationEventType.Move)
                    OverlappingMoved(this, args);
                if (type == RelationEventType.End)
                    OverlappingEnded(this, args);
                Overlapped(this, args);
            }
        }

        private OverlappingElementsRelationEventArgs CreateEventArgs(RelationEventType eventType, Overlap overlapping)
        {
            OverlappingElementsRelationEventArgs args = new OverlappingElementsRelationEventArgs(overlapping);
            SetDefaultEventArgs(args, eventType);
            args.Element1 = element1;
            args.Element2 = element2;
            if (overlapping != null)
            {
                args.ElementArgs1 = new ElementArgs(element1, overlapping);
                args.ElementArgs2 = new ElementArgs(element2, overlapping);
            }
            return args;
        }
    }
}
