﻿using System.Windows.Media.Media3D;
using ProjectorToolkit.Elements;
using ProjectorToolkit.Utilities;
using ProjectorToolkit.World;

namespace ProjectorToolkit.Relations
{
    public class PointingRelationEventArgs : RelationEventArgs
    {
        public AbstractElement Pointer;
        public AbstractElement Target;
        public Point3D Intersection;
        public Vector3D IntersectionFromCenter;
        public Angle PointerRotation;
        public double Distance;
    }

    /// <summary>
    /// This relation triggers callbacks when two elements are pointing to each other.
    /// </summary>
    public sealed class PointingRelation : AbstractRelation<PointingRelationEventArgs>
    {
        public delegate void PointingRelationEventHandler(object sender, PointingRelationEventArgs e);
        public event PointingRelationEventHandler Pointed = delegate { };
        public event PointingRelationEventHandler PointingStarted = delegate { };
        public event PointingRelationEventHandler PointingMoved = delegate { };
        public event PointingRelationEventHandler PointingEnded = delegate { };

        private AbstractElement pointer;
        private AbstractElement target;
        private bool isInUpdate = false;

        public PointingRelation(AbstractElement pointer, AbstractElement target)
        {
            this.pointer = pointer;
            this.target = target;

            pointer.Updated += NoRecursiveUpdate;
            target.Updated += NoRecursiveUpdate;
        }

        protected override void Update(AbstractElement element)
        {
            AbstractElement target = ServerWorld.Instance.PointingTarget(pointer);

            RelationEventType? type = SetRelationType(this.target == target);
            if (type != null)
            {
                PointingRelationEventArgs args = CreateEventArgs((RelationEventType)type);

                // Call the callbacks
                if (type == RelationEventType.Start)
                    PointingStarted(this, args);
                if (type == RelationEventType.Move)
                    PointingMoved(this, args);
                if (type == RelationEventType.End)
                    PointingEnded(this, args);
                Pointed(this, args);
            }
        }

        private PointingRelationEventArgs CreateEventArgs(RelationEventType eventType)
        {
            PointingRelationEventArgs args = new PointingRelationEventArgs();
            SetDefaultEventArgs(args, eventType);
            args.Pointer = pointer;
            args.Target = target;

            Plane plane = new Plane(pointer.Position, pointer.Direction);
            args.PointerRotation = Angle.Between(pointer.Up, plane.Up);

            Point3D? hitpoint = target.Intersection(pointer);
            if (hitpoint != null)
            {
                args.Intersection = (Point3D)hitpoint;
                args.IntersectionFromCenter = (Point3D)hitpoint - target.Position;
                args.Distance = pointer.Position.DistanceTo(args.Intersection);
            }
            return args;
        }
    }
}
