﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Media3D;
using System.Windows.Shapes;
using System.Windows.Threading;
using ProjectorToolkit.Elements;
using ProjectorToolkit.Utilities;
using ProjectorToolkit.World;
using System.Diagnostics;

namespace ProjectorToolkit.UI
{
    public class Viewport : Grid
    {
        #region Events
        public delegate void DisplayFoundHandler(AbstractDisplay display);
        private DisplayFoundHandler DisplayFound = delegate { };
        public event DisplayFoundHandler DisplayFoundCallback
        {
            add
            {
                DisplayFound += value;
                if (display != null)
                    DisplayFound(display);
            }
            remove { DisplayFound -= value; }
        }
        #endregion

        //TODO Save elements and names
        private List<string> blackoutNames = new List<string>();
        private List<string> hiddenElementNames = new List<string>();
        private string viewportName;
        private AbstractDisplay display;
        private ClientWorld world;
        private Dictionary<AbstractDisplay, Rectangle> cachedRectangles = new Dictionary<AbstractDisplay, Rectangle>();

        private Viewport3D xaml_viewport;
        private Model3DGroup xaml_geometry;
        private Grid xaml_clipping_grid;

        #region Properties
        public Model3DGroup Model3DGroup
        {
            get { return xaml_geometry; }
        }
        public ClientWorld World
        {
            get { return world; }
            set { world = value; }
        }
        public AbstractDisplay Display
        {
            get { return display; }
        }
        #endregion

        public Viewport()
        {
            this.Background = new SolidColorBrush(Colors.Black);

            xaml_viewport = new Viewport3D();
            xaml_viewport.IsHitTestVisible = false;
            xaml_viewport.ClipToBounds = false;
            xaml_viewport.HorizontalAlignment = System.Windows.HorizontalAlignment.Stretch;
            xaml_viewport.VerticalAlignment = System.Windows.VerticalAlignment.Stretch;

            // Add an ambient light to the world
            ModelVisual3D visual3D = new ModelVisual3D();
            xaml_geometry = new Model3DGroup();
            xaml_geometry.Children.Add(new AmbientLight(Colors.White));
            visual3D.Content = xaml_geometry;
            xaml_viewport.Children.Add(visual3D);

            // Create grid that contains clipping rectangles
            xaml_clipping_grid = new Grid();
            xaml_clipping_grid.HorizontalAlignment = System.Windows.HorizontalAlignment.Stretch;
            xaml_clipping_grid.VerticalAlignment = System.Windows.VerticalAlignment.Stretch;

            this.Children.Add(xaml_viewport);
            this.Children.Add(xaml_clipping_grid);
        }

        /// <summary>
        /// Start the client with the given viewport.
        /// A world has to be set before starting the client.
        /// </summary>
        /// <param name="viewportName">Name of the viewport in the world.</param>
        public void Start(string viewportName)
        {
            if (world == null)
                throw new Exception("Set a world before starting the client!");

            this.display = null;
            this.viewportName = viewportName;

            world.Updated += WorldUpdate;
        }

        private bool IsHidden(AbstractElement element)
        {
            string name = world.GetNameOf(element);
            return (display != null && display.HiddenElements.Contains(name));
        }

        private void WorldUpdate(object sender, UpdateEventEventArgs e)
        {
            switch (e.ElementState)
            {
                case ElementState.New:
                    if (!IsHidden(e.Element))
                        ShowElement(e.Element);
                    break;
                case ElementState.Focus:
                    if (!IsHidden(e.Element))
                        FocusElement(e.Element);
                    break;
                case ElementState.Gone:
                    HideElement(e.Element);
                    break;
            }

            AbstractElement element;
            if (display == null && world.TryGetElement(viewportName, out element))
            {
                display = (AbstractDisplay)element;
                display.Updated += UpdateViewport;
                DisplayFound(display);
            }
            if (display != null)
                UpdateViewport(display);
        }

        private void UpdateViewport(AbstractElement element)
        {
            xaml_geometry.Dispatcher.Invoke(DispatcherPriority.Normal, new Action(delegate()
            {
                UpdateBlackouts(display);
                UpdateHiddenElements(display);
                xaml_viewport.Camera = display.Camera;
            }));
        }

        private void UpdateHiddenElements(AbstractDisplay display)
        {
            foreach (KeyValuePair<string, AbstractElement> ne in world.NameAndElements)
            {
                if (display.HiddenElements.Contains(ne.Key))
                    HideElement(ne.Value);
                else
                    ShowElement(ne.Value);
            }
        }

        private void UpdateBlackouts(AbstractDisplay e)
        {
            List<string> blackoutList = new List<string>(display.Blackouts.Keys);
            // Unregister old overlappings
            IEnumerable<string> oldBlackouts = blackoutNames.Except(blackoutList);
            foreach (string elementName in oldBlackouts)
            {
                AbstractElement element;
                if (world.TryGetElement(elementName, out element))
                {
                    if (element is AbstractDisplay)
                    {
                        element.Updated -= UpdateClipping;
                        if (cachedRectangles.ContainsKey((AbstractDisplay)element) && xaml_clipping_grid.Children.Contains(cachedRectangles[(AbstractDisplay)element]))
                            xaml_clipping_grid.Children.Remove(cachedRectangles[(AbstractDisplay)element]);
                    }
                    if (element is AbstractVirtualElement2D)
                        ((AbstractVirtualElement2D)element).Blackout = false;
                }
            }

            // Register new overlappings
            IEnumerable<string> newBlackouts = blackoutList.Except(blackoutNames);
            foreach (string elementName in newBlackouts)
            {
                AbstractElement element;
                if (world.TryGetElement(elementName, out element))
                {
                    if (element is AbstractDisplay)
                    {
                        element.Updated += UpdateClipping;
                        UpdateClipping(element);
                        if (cachedRectangles.ContainsKey((AbstractDisplay)element) && !xaml_clipping_grid.Children.Contains(cachedRectangles[(AbstractDisplay)element]))
                            xaml_clipping_grid.Children.Add(cachedRectangles[(AbstractDisplay)element]);
                    }
                    if (element is AbstractVirtualElement2D)
                        ((AbstractVirtualElement2D)element).Blackout = true;
                }
            }
            blackoutNames = blackoutList;

            // Update clippings
            AbstractElement elem;
            foreach (string name in blackoutNames)
            {
                if (world.TryGetElement(name, out elem) && elem is AbstractDisplay)
                    UpdateClipping(elem);
            }
        }

        private Rectangle CreateRectangle()
        {
            Rectangle rectangle = new Rectangle();
            rectangle.HorizontalAlignment = System.Windows.HorizontalAlignment.Stretch;
            rectangle.VerticalAlignment = System.Windows.VerticalAlignment.Stretch;
            rectangle.Stroke = new SolidColorBrush(Colors.Black);
            rectangle.Fill = new SolidColorBrush(Colors.Black);
            return rectangle;
        }

        /// <summary>
        /// Updates the rectangle clippings of the given display.
        /// </summary>
        /// <param name="e">Display to update.</param>
        private void UpdateClipping(AbstractElement e)
        {
            xaml_geometry.Dispatcher.Invoke(DispatcherPriority.Normal, new Action(delegate()
            {
                string elementName;

                if (world.TryGetName(e, out elementName) && e is AbstractDisplay)
                {
                    AbstractDisplay vp = (AbstractDisplay)e;

                    Overlap globalOverlaps;
                    if (vp is OrthographicDisplay)
                        globalOverlaps = Overlap.Of(display, (IPositionable2D)vp);
                    else if (display is OrthographicDisplay)
                        globalOverlaps = Overlap.Of(vp, (IPositionable2D)display);
                    else
                        globalOverlaps = Overlap.Of(display, vp);

                    Debug.Assert(globalOverlaps.Overlaps.Count <= 1);

                    // Only one overlap possible for convex polygons
                    if (globalOverlaps != null)
                    {
                        Point3D[] points = globalOverlaps.PolygonToAbsolute(globalOverlaps.Overlaps[0]);
                        Point[] localOverlap = display.ConvertToPixels(points);

                        // Create rectangles only once and reuse old ones
                        Rectangle rectangle;
                        if (!cachedRectangles.TryGetValue(vp, out rectangle))
                        {
                            rectangle = CreateRectangle();
                            cachedRectangles[vp] = rectangle;
                            PathGeometry pg = new PathGeometry();
                            pg.Figures.Add(new PathFigure());
                            rectangle.Clip = pg;
                        }

                        // Manipulate clippable figure 
                        PathFigure pf = ((PathGeometry)rectangle.Clip).Figures[0];

                        // Adjust size of collection
                        while (pf.Segments.Count < localOverlap.Length)
                            pf.Segments.Add(new System.Windows.Media.LineSegment());
                        while (pf.Segments.Count > localOverlap.Length)
                            pf.Segments.RemoveAt(pf.Segments.Count - 1);
                        pf.StartPoint = localOverlap[0];

                        // Manipulate collection
                        for (int i = 0; i < localOverlap.Length; i++)
                            ((System.Windows.Media.LineSegment)pf.Segments[i]).Point = localOverlap[i];

                        rectangle.Opacity = display.Blackouts[elementName];
                    }
                }
            }));
        }

        #region Element Handling
        public void FocusElement(AbstractElement element)
        {
            if (world.Contains(element))
            {
                HideElement(element);
                ShowElement(element);
            }
        }

        protected void ShowElement(AbstractElement element)
        {
            Model3DGroup.Dispatcher.Invoke(DispatcherPriority.Normal, new Action(delegate()
            {
                if (element.Mesh != null && !Model3DGroup.Children.Contains(element.Mesh))
                {
                    Model3DGroup.Children.Add(element.Mesh);
                    element.UpdateMesh();
                }
            }));
        }

        protected void HideElement(AbstractElement element)
        {
            Model3DGroup.Dispatcher.Invoke(DispatcherPriority.Normal, new Action(delegate()
            {
                if (element.Mesh != null && Model3DGroup.Children.Contains(element.Mesh))
                    Model3DGroup.Children.Remove(element.Mesh);
            }));
        }
        #endregion
    }
}
