﻿using System;
using System.Windows.Media.Media3D;

namespace ProjectorToolkit.Utilities
{
    public static class Vector3DExtensions
    {
        public static Point3D ToPoint3D(this Vector3D vector)
        {
            return new Point3D(vector.X, vector.Y, vector.Z);
        }

        /// <summary>
        /// Perform the same rotation on the given vector as happend between old and new vector.
        /// </summary>
        /// <param name="vector">Vector to perform rotation on</param>
        /// <param name="oldVector">Old vector of the rotation</param>
        /// <param name="newVector">New vector of the rotation</param>
        /// <returns>The rotated vector</returns>
        public static Vector3D SameRotation(Vector3D vector, Vector3D oldVector, Vector3D newVector)
        {
            oldVector.Normalize();
            newVector.Normalize();
            Vector3D cross = Vector3D.CrossProduct(oldVector, newVector);
            Angle angle = Angle.Between(oldVector, newVector);
            return RotationAroundArbitraryAxis(vector, new Point3D(0, 0, 0), cross, angle);
        }

        #region Point3D Rotation
        public static Vector3D RotationAroundArbitraryAxis(Vector3D vector, Point3D axisPoint, Vector3D axisDirection, Angle angle)
        {
            Point3D p = RotationAroundArbitraryAxis(new Point3D(vector.X, vector.Y, vector.Z), axisPoint, axisDirection, angle);
            return p.ToVector3D();
        }
        public static Point3D RotationAroundArbitraryAxis(Point3D point, Point3D axisPoint, Vector3D axisDirection, Angle angle)
        {
            if (angle.IsNotZero)
            {
                double radian = angle.InRadians;
                double cosAngle = Math.Cos(radian);
                double sinAngle = Math.Sin(radian);
                double oneMinusCosAngle = (1.0 - cosAngle);

                Vector3D aD = axisDirection;
                aD.Normalize();
                Point3D aP = axisPoint;
                double uSqr = Math.Pow(aD.X, 2);
                double vSqr = Math.Pow(aD.Y, 2);
                double wSqr = Math.Pow(aD.Z, 2);
                double minusUxVyWz = -aD.X * point.X - aD.Y * point.Y - aD.Z * point.Z;

                double firstX = aP.X * (vSqr + wSqr) - aD.X * (aP.Y * aD.Y + aP.Z * aD.Z + minusUxVyWz);
                double firstY = aP.Y * (uSqr + wSqr) - aD.Y * (aP.X * aD.X + aP.Z * aD.Z + minusUxVyWz);
                double firstZ = aP.Z * (uSqr + vSqr) - aD.Z * (aP.X * aD.X + aP.Y * aD.Y + minusUxVyWz);

                double secondX = (-aP.Z * aD.Y + aP.Y * aD.Z - aD.Z * point.Y + aD.Y * point.Z);
                double secondY = (+aP.Z * aD.X - aP.X * aD.Z + aD.Z * point.X - aD.X * point.Z);
                double secondZ = (-aP.Y * aD.X + aP.X * aD.Y - aD.Y * point.X + aD.X * point.Y);

                return new Point3D(
                     firstX * oneMinusCosAngle + (point.X * cosAngle) + secondX * sinAngle,
                     firstY * oneMinusCosAngle + (point.Y * cosAngle) + secondY * sinAngle,
                     firstZ * oneMinusCosAngle + (point.Z * cosAngle) + secondZ * sinAngle
                    );
            }
            else
                return point;
        }

        public static Point3D RotationAroundXAxis(Point3D point, Angle angle)
        {
            if (angle.IsNotZero)
            {
                double radian = angle.InRadians;
                return new Point3D(
                    point.X,
                    point.Y * Math.Cos(radian) - point.Z * Math.Sin(radian),
                    point.Y * Math.Sin(radian) + point.Z * Math.Cos(radian));
            }
            else
                return point;
        }

        public static Point3D RotationAroundYAxis(Point3D point, Angle angle)
        {
            if (angle.IsNotZero)
            {
                double radian = angle.InRadians;
                return new Point3D(
                    point.Z * Math.Sin(radian) + point.X * Math.Cos(radian),
                    point.Y,
                    point.Z * Math.Cos(radian) - point.X * Math.Sin(radian));
            }
            else
                return point;
        }

        public static Point3D RotationAroundZAxis(Point3D point, Angle angle)
        {
            if (angle.IsNotZero)
            {
                double radian = angle.InRadians;
                return new Point3D(
                    point.X * Math.Cos(radian) - point.Y * Math.Sin(radian),
                    point.X * Math.Sin(radian) + point.Y * Math.Cos(radian),
                    point.Z);
            }
            else
                return point;
        }
        #endregion
    }
}
